import threading
from typing import Optional

from pydantic import BaseModel

from prowler.lib.logger import logger
from prowler.lib.scan_filters.scan_filters import is_resource_filtered
from prowler.providers.aws.aws_provider import generate_regional_clients


################################ WellArchitected
class WellArchitected:
    def __init__(self, audit_info):
        self.service = "wellarchitected"
        self.session = audit_info.audit_session
        self.audit_resources = audit_info.audit_resources
        self.regional_clients = generate_regional_clients(self.service, audit_info)
        self.workloads = []
        self.__threading_call__(self.__list_workloads__)
        self.__list_tags_for_resource__()

    def __get_session__(self):
        return self.session

    def __threading_call__(self, call):
        threads = []
        for regional_client in self.regional_clients.values():
            threads.append(threading.Thread(target=call, args=(regional_client,)))
        for t in threads:
            t.start()
        for t in threads:
            t.join()

    def __list_workloads__(self, regional_client):
        logger.info("WellArchitected - Listing Workloads...")
        try:
            for workload in regional_client.list_workloads()["WorkloadSummaries"]:
                if not self.audit_resources or (
                    is_resource_filtered(workload["WorkloadArn"], self.audit_resources)
                ):
                    self.workloads.append(
                        Workload(
                            id=workload["WorkloadId"],
                            arn=workload["WorkloadArn"],
                            name=workload["WorkloadName"],
                            region=regional_client.region,
                            lenses=workload["Lenses"],
                            improvement_status=workload["ImprovementStatus"],
                            risks=workload["RiskCounts"],
                        )
                    )

        except Exception as error:
            logger.error(
                f"{regional_client.region} -- {error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
            )

    def __list_tags_for_resource__(self):
        logger.info("WellArchitected - Listing Tags...")
        try:
            for workload in self.workloads:
                regional_client = self.regional_clients[workload.region]
                response = regional_client.list_tags_for_resource(
                    WorkloadArn=workload.arn
                )["Tags"]
                workload.tags = [response]
        except Exception as error:
            logger.error(
                f"{regional_client.region} -- {error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
            )


class Workload(BaseModel):
    id: str
    arn: str
    name: str
    region: str
    lenses: list
    improvement_status: str
    risks: dict
    tags: Optional[list] = []



from pydantic import BaseModel

from prowler.lib.logger import logger
from prowler.providers.gcp.gcp_provider import generate_client
from prowler.providers.gcp.services.compute.compute_client import compute_client


################## Dataproc
class Dataproc:
    def __init__(self, audit_info):
        self.service = "dataproc"
        self.api_version = "v1"
        self.project_ids = audit_info.project_ids
        self.default_project_id = audit_info.default_project_id
        self.client = generate_client(self.service, self.api_version, audit_info)
        self.clusters = []
        self.__get_clusters__()

    def __get_clusters__(self):
        for project_id in self.project_ids:
            try:
                for region in compute_client.regions:
                    request = (
                        self.client.projects()
                        .regions()
                        .clusters()
                        .list(projectId=project_id, region=region)
                    )
                    while request is not None:
                        response = request.execute()

                        for cluster in response.get("clusters", []):
                            self.clusters.append(
                                Cluster(
                                    name=cluster["clusterName"],
                                    id=cluster["clusterUuid"],
                                    encryption_config=cluster["config"][
                                        "encryptionConfig"
                                    ],
                                    project_id=project_id,
                                )
                            )

                        request = (
                            self.client.projects()
                            .regions()
                            .clusters()
                            .list_next(
                                previous_request=request, previous_response=response
                            )
                        )
            except Exception as error:
                logger.error(
                    f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
                )


class Cluster(BaseModel):
    name: str
    id: str
    encryption_config: dict
    project_id: str



from pydantic import BaseModel

from prowler.lib.logger import logger
from prowler.providers.gcp.gcp_provider import generate_client


################## CloudResourceManager
class CloudResourceManager:
    def __init__(self, audit_info):
        self.service = "cloudresourcemanager"
        self.api_version = "v1"
        self.region = "global"
        self.project_ids = audit_info.project_ids
        self.client = generate_client(self.service, self.api_version, audit_info)
        self.bindings = []
        self.__get_iam_policy__()

    def __get_client__(self):
        return self.client

    def __get_iam_policy__(self):
        for project_id in self.project_ids:
            try:
                policy = (
                    self.client.projects().getIamPolicy(resource=project_id).execute()
                )
                for binding in policy["bindings"]:
                    self.bindings.append(
                        Binding(
                            role=binding["role"],
                            members=binding["members"],
                            project_id=project_id,
                        )
                    )
            except Exception as error:
                logger.error(
                    f"{self.region} -- {error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
                )


class Binding(BaseModel):
    role: str
    members: list
    project_id: str



from pydantic import BaseModel

from prowler.lib.logger import logger
from prowler.providers.gcp.gcp_provider import generate_client


################## DNS
class DNS:
    def __init__(self, audit_info):
        self.service = "dns"
        self.api_version = "v1"
        self.project_ids = audit_info.project_ids
        self.default_project_id = audit_info.default_project_id
        self.client = generate_client(self.service, self.api_version, audit_info)
        self.managed_zones = []
        self.__get_managed_zones__()

    def __get_managed_zones__(self):
        for project_id in self.project_ids:
            try:
                request = self.client.managedZones().list(project=project_id)
                while request is not None:
                    response = request.execute()
                    for managed_zone in response.get("managedZones"):
                        self.managed_zones.append(
                            ManagedZone(
                                name=managed_zone["name"],
                                id=managed_zone["id"],
                                dnssec=managed_zone["dnssecConfig"]["state"] == "on",
                                key_specs=managed_zone["dnssecConfig"][
                                    "defaultKeySpecs"
                                ],
                                project_id=project_id,
                            )
                        )

                    request = self.client.managedZones().list_next(
                        previous_request=request, previous_response=response
                    )
            except Exception as error:
                logger.error(
                    f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
                )


class ManagedZone(BaseModel):
    name: str
    id: str
    dnssec: bool
    key_specs: list
    project_id: str



from pydantic import BaseModel

from prowler.lib.logger import logger
from prowler.providers.gcp.gcp_provider import generate_client


################## Compute
class Compute:
    def __init__(self, audit_info):
        self.service = "compute"
        self.api_version = "v1"
        self.project_ids = audit_info.project_ids
        self.default_project_id = audit_info.default_project_id
        self.client = generate_client(self.service, self.api_version, audit_info)
        self.regions = set()
        self.zones = set()
        self.instances = []
        self.networks = []
        self.firewalls = []
        self.__get_regions__()
        self.__get_zones__()
        self.__get_instances__()
        self.__get_networks__()
        self.__get_firewalls__()

    def __get_regions__(self):
        for project_id in self.project_ids:
            try:
                request = self.client.regions().list(project=project_id)
                while request is not None:
                    response = request.execute()

                    for region in response.get("items", []):
                        self.regions.add(region["name"])

                    request = self.client.regions().list_next(
                        previous_request=request, previous_response=response
                    )
            except Exception as error:
                logger.error(
                    f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
                )

    def __get_zones__(self):
        for project_id in self.project_ids:
            try:
                request = self.client.zones().list(project=project_id)
                while request is not None:
                    response = request.execute()

                    for zone in response.get("items", []):
                        self.zones.add(zone["name"])

                    request = self.client.zones().list_next(
                        previous_request=request, previous_response=response
                    )
            except Exception as error:
                logger.error(
                    f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
                )

    def __get_instances__(self):
        for project_id in self.project_ids:
            try:
                for zone in self.zones:
                    request = self.client.instances().list(
                        project=project_id, zone=zone
                    )
                    while request is not None:
                        response = request.execute()

                        for instance in response.get("items", []):
                            public_ip = False
                            for interface in instance["networkInterfaces"]:
                                for config in interface.get("accessConfigs", []):
                                    if "natIP" in config:
                                        public_ip = True
                            self.instances.append(
                                Instance(
                                    name=instance["name"],
                                    id=instance["id"],
                                    zone=zone,
                                    public_ip=public_ip,
                                    metadata=instance["metadata"],
                                    shielded_enabled_vtpm=instance[
                                        "shieldedInstanceConfig"
                                    ]["enableVtpm"],
                                    shielded_enabled_integrity_monitoring=instance[
                                        "shieldedInstanceConfig"
                                    ]["enableIntegrityMonitoring"],
                                    service_accounts=instance["serviceAccounts"],
                                    ip_forward=instance.get("canIpForward", False),
                                    disks_encryption=[
                                        (
                                            disk["deviceName"],
                                            True
                                            if disk.get("diskEncryptionKey", {}).get(
                                                "sha256"
                                            )
                                            else False,
                                        )
                                        for disk in instance["disks"]
                                    ],
                                    project_id=project_id,
                                )
                            )

                        request = self.client.instances().list_next(
                            previous_request=request, previous_response=response
                        )
            except Exception as error:
                logger.error(
                    f"{zone} -- {error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
                )

    def __get_networks__(self):
        for project_id in self.project_ids:
            try:
                request = self.client.networks().list(project=project_id)
                while request is not None:
                    response = request.execute()
                    for network in response.get("items", []):
                        self.networks.append(
                            Network(
                                name=network["name"],
                                id=network["id"],
                                project_id=project_id,
                            )
                        )

                    request = self.client.networks().list_next(
                        previous_request=request, previous_response=response
                    )
            except Exception as error:
                logger.error(
                    f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
                )

    def __get_firewalls__(self):
        for project_id in self.project_ids:
            try:
                request = self.client.firewalls().list(project=project_id)
                while request is not None:
                    response = request.execute()

                    for firewall in response.get("items", []):
                        self.firewalls.append(
                            Firewall(
                                name=firewall["name"],
                                id=firewall["id"],
                                source_ranges=firewall["sourceRanges"],
                                direction=firewall["direction"],
                                allowed_rules=firewall.get("allowed", []),
                                project_id=project_id,
                            )
                        )

                    request = self.client.firewalls().list_next(
                        previous_request=request, previous_response=response
                    )
            except Exception as error:
                logger.error(
                    f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
                )


class Instance(BaseModel):
    name: str
    id: str
    zone: str
    public_ip: bool
    project_id: str
    metadata: dict
    shielded_enabled_vtpm: bool
    shielded_enabled_integrity_monitoring: bool
    service_accounts: list
    ip_forward: bool
    disks_encryption: list


class Network(BaseModel):
    name: str
    id: str
    project_id: str


class Firewall(BaseModel):
    name: str
    id: str
    source_ranges: list
    direction: str
    allowed_rules: list
    project_id: str



import os
import sys

from prowler.config.config import (
    json_asff_file_suffix,
    json_file_suffix,
    json_ocsf_file_suffix,
    prowler_version,
    timestamp,
    timestamp_utc,
)
from prowler.lib.logger import logger
from prowler.lib.outputs.models import (
    Account,
    Check_Output_JSON_OCSF,
    Cloud,
    Compliance,
    Compliance_OCSF,
    Feature,
    Finding,
    Group,
    Metadata,
    Organization,
    Product,
    ProductFields,
    Remediation_OCSF,
    Resource,
    Resources,
    Severity,
    get_check_compliance,
    unroll_dict_to_list,
)
from prowler.lib.utils.utils import hash_sha512, open_file


def fill_json_asff(finding_output, audit_info, finding, output_options):
    # Check if there are no resources in the finding
    if finding.resource_arn == "":
        if finding.resource_id == "":
            finding.resource_id = "NONE_PROVIDED"
        finding.resource_arn = finding.resource_id
    finding_output.Id = f"prowler-{finding.check_metadata.CheckID}-{audit_info.audited_account}-{finding.region}-{hash_sha512(finding.resource_id)}"
    finding_output.ProductArn = f"arn:{audit_info.audited_partition}:securityhub:{finding.region}::product/prowler/prowler"
    finding_output.ProductFields = ProductFields(
        ProviderVersion=prowler_version, ProwlerResourceName=finding.resource_arn
    )
    finding_output.GeneratorId = "prowler-" + finding.check_metadata.CheckID
    finding_output.AwsAccountId = audit_info.audited_account
    finding_output.Types = finding.check_metadata.CheckType
    finding_output.FirstObservedAt = (
        finding_output.UpdatedAt
    ) = finding_output.CreatedAt = timestamp_utc.strftime("%Y-%m-%dT%H:%M:%SZ")
    finding_output.Severity = Severity(Label=finding.check_metadata.Severity.upper())
    finding_output.Title = finding.check_metadata.CheckTitle
    finding_output.Description = finding.status_extended
    finding_output.Resources = [
        Resource(
            Id=finding.resource_arn,
            Type=finding.check_metadata.ResourceType,
            Partition=audit_info.audited_partition,
            Region=finding.region,
        )
    ]
    # Iterate for each compliance framework
    compliance_summary = []
    associated_standards = []
    check_compliance = get_check_compliance(finding, "aws", output_options)
    for key, value in check_compliance.items():
        associated_standards.append({"StandardsId": key})
        item = f"{key} {' '.join(value)}"
        if len(item) > 64:
            item = item[0:63]
        compliance_summary.append(item)

    # Add ED to PASS or FAIL (PASSED/FAILED)
    finding_output.Compliance = Compliance(
        Status=finding.status + "ED",
        AssociatedStandards=associated_standards,
        RelatedRequirements=compliance_summary,
    )
    finding_output.Remediation = {
        "Recommendation": finding.check_metadata.Remediation.Recommendation
    }

    return finding_output


def fill_json_ocsf(
    finding_output: Check_Output_JSON_OCSF, audit_info, finding, output_options
):
    resource_region = ""
    resource_name = ""
    resource_uid = ""
    finding_uid = ""
    resource_labels = finding.resource_tags if finding.resource_tags else []
    if finding.status == "PASS":
        finding_output.status = "Success"
        finding_output.status_id = 1
    elif finding.status == "FAIL":
        finding_output.status = "Failure"
        finding_output.status_id = 2
    finding_output.status_detail = finding_output.message = finding.status_extended
    finding_output.severity = finding.check_metadata.Severity
    if finding_output.severity == "low":
        finding_output.severity_id = 2
    elif finding_output.severity == "medium":
        finding_output.severity_id = 3
    elif finding_output.severity == "high":
        finding_output.severity_id = 4
    elif finding_output.severity == "critical":
        finding_output.severity_id = 5
    aws_account_name = ""
    aws_org_uid = ""
    if (
        hasattr(audit_info, "organizations_metadata")
        and audit_info.organizations_metadata
    ):
        aws_account_name = audit_info.organizations_metadata.account_details_name
        aws_org_uid = audit_info.organizations_metadata.account_details_org
    finding_output.cloud = Cloud(
        provider=finding.check_metadata.Provider,
    )
    if finding.check_metadata.Provider == "aws":
        finding_output.cloud.account = Account(
            name=aws_account_name,
            uid=audit_info.audited_account,
        )
        finding_output.cloud.org = Organization(
            name=aws_org_uid,
            uid=aws_org_uid,
        )
        finding_output.cloud.region = resource_region = finding.region
        resource_name = finding.resource_id
        resource_uid = finding.resource_arn
        finding_uid = f"prowler-{finding.check_metadata.Provider}-{finding.check_metadata.CheckID}-{audit_info.audited_account}-{finding.region}-{finding.resource_id}"
    elif finding.check_metadata.Provider == "azure":
        finding_output.cloud.account = Account(
            name=finding.subscription,
            uid=finding.subscription,
        )
        finding_output.cloud.org = Organization(
            name=audit_info.identity.domain,
            uid=audit_info.identity.domain,
        )
        resource_name = finding.resource_name
        resource_uid = finding.resource_id
        finding_uid = f"prowler-{finding.check_metadata.Provider}-{finding.check_metadata.CheckID}-{finding.subscription}-{finding.resource_id}"
    elif finding.check_metadata.Provider == "gcp":
        finding_output.cloud.account = None
        finding_output.cloud.org = None
        finding_output.cloud.project_uid = finding.project_id
        finding_output.cloud.region = resource_region = finding.location
        resource_name = finding.resource_name
        resource_uid = finding.resource_id
        finding_uid = f"prowler-{finding.check_metadata.Provider}-{finding.check_metadata.CheckID}-{finding.project_id}-{finding.resource_id}"
    finding_output.finding = Finding(
        title=finding.check_metadata.CheckTitle,
        uid=finding_uid,
        desc=finding.check_metadata.Description,
        supporting_data={
            "Risk": finding.check_metadata.Risk,
            "Notes": finding.check_metadata.Notes,
        },
        related_events=finding.check_metadata.DependsOn
        + finding.check_metadata.RelatedTo,
        remediation=Remediation_OCSF(
            kb_articles=list(
                filter(
                    None,
                    [
                        finding.check_metadata.Remediation.Code.NativeIaC,
                        finding.check_metadata.Remediation.Code.Terraform,
                        finding.check_metadata.Remediation.Code.CLI,
                        finding.check_metadata.Remediation.Code.Other,
                        finding.check_metadata.Remediation.Recommendation.Url,
                    ],
                )
            ),
            desc=finding.check_metadata.Remediation.Recommendation.Text,
        ),
        types=finding.check_metadata.CheckType,
        src_url=finding.check_metadata.RelatedUrl,
    )
    finding_output.resources.append(
        Resources(
            group=Group(name=finding.check_metadata.ServiceName),
            region=resource_region,
            name=resource_name,
            labels=resource_labels,
            uid=resource_uid,
            type=finding.check_metadata.ResourceType,
            details=finding.resource_details,
        )
    )
    finding_output.time = timestamp.isoformat()
    finding_output.metadata = Metadata(
        product=Product(
            feature=Feature(
                uid=finding.check_metadata.CheckID,
                name=finding.check_metadata.CheckID,
            )
        ),
        original_time=timestamp.isoformat(),
        profiles=[audit_info.profile]
        if hasattr(audit_info, "organizations_metadata")
        else [],
    )
    finding_output.compliance = Compliance_OCSF(
        status=finding_output.status,
        status_detail=finding_output.status_detail,
        requirements=unroll_dict_to_list(
            get_check_compliance(
                finding, finding.check_metadata.Provider, output_options
            )
        ),
    )

    return finding_output


def close_json(output_filename, output_directory, mode):
    """close_json closes the output JSON file replacing the last comma with ]"""
    try:
        suffix = json_file_suffix
        if mode == "json-asff":
            suffix = json_asff_file_suffix
        elif mode == "json-ocsf":
            suffix = json_ocsf_file_suffix
        filename = f"{output_directory}/{output_filename}{suffix}"
        # Close JSON file if exists
        if os.path.isfile(filename):
            file_descriptor = open_file(
                filename,
                "a",
            )
            # Replace last comma for square bracket if not empty
            if file_descriptor.tell() > 0:
                if file_descriptor.tell() != 1:
                    file_descriptor.seek(file_descriptor.tell() - 1, os.SEEK_SET)
                file_descriptor.truncate()
                file_descriptor.write("]")
            file_descriptor.close()
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}] -- {error}"
        )
        sys.exit(1)
